/********************************************************
 * File: smlex.c (Listing 7)
 * Title: Lexical Analyzer
 * Author: Allen Holub
 * Compiler: Turbo C 2.0, MSC 5.1
 ********************************************************/

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "smtok.h"

/* SMLEX.C:  This file is the lexical analyzer for the parser in smparse.c.
 *           it recognizes the tokens defined in smtok.h. Public symbols are:
 *
 * gtok( FILE *fp );   Return next token from the indicated input file.
 * char yytext[];      Lexeme for token returned from most recent gtok() call.
 * int  yylineno;      Input line number.
 */

extern void error ( char *fmt, ... );        /* source in smparse.c */

/*----------------------------------------------------------------------*/
char yytext[10 * 1024]; /* lexeme               */
int  yylineno;          /* input line number    */

#define         STACKSIZE 128
static int      Stack[STACKSIZE];               /* stack for pushb      */
static int      *Sp = Stack + STACKSIZE;        /* stack pointer        */

/*----------------------------------------------------------------------*/
/* get():   If there's something on the stack, return it as the next input
 *          character, otherwise call getc().
 */

static int get( FILE *fp )
{
    int c = (Sp >= Stack + STACKSIZE) ? getc(fp) : *Sp++;
    if( c == '\n')
        ++yylineno;
    return c;
}

static void push( int c )         /* push back a single input character */
{
    if( c == '\n' )
        --yylineno;

    if( Sp == Stack )
    {
        error( "Pushback-stack overflow\n" );
        exit(1);
    }
    *--Sp = c;
}

static void pushb( int n, char *buf )   /* Push all but the first n charac- */
{                                       /* ters of buf back into the input. */

    char *end;

    buf += n;
    end = buf + strlen(buf);

    while( --end >= buf )
        push( *end );
}

/*----------------------------------------------------------------------*/

static int cmp( char *substr, char *str )
{
    /* Like strcmp, but substr can be shorter than str. Match is successful
     * if you get to the end of substr without a mismatch.
     */

    for(; *str==*substr && *substr; ++str, ++substr )
        ;
    return (!*substr) ? 0 : (*str - *substr);
}

/*----------------------------------------------------------------------*/

int gtok( FILE *input )                 /* return next input token */
{
    /* A brute-force lexical analyzer. Demonstrates, through its ugliness,
     * why a state-machine approach is worthwhile.
     */

    static int  lookahead = -1; /* next input character, -1 to force read  */
    int         rval;           /* return value.                           */
    int         level;          /* brace nesting level for CODE processing */
    char        *p;
    int         i;

    if( lookahead == -1 )
        if( (lookahead = get(input)) == -1 )
            return END_OF_FILE;

    /* Ignore whitespace and comments. */

    while( isspace(lookahead) || lookahead == '/' )
    {
        if( lookahead == '/' )
        {
            if( (lookahead = getc(input)) == '/' )
                while( (lookahead = getc(input)) != '\n' && lookahead != EOF )
                    ;
            else                        /* Character following / wasn't  */
            {                           /* another slash. push it back   */
                push( lookahead );      /* into the input stream and     */
                lookahead = '/';        /* proceeded with the first slash */
                break;                  /* as the lookahead character.   */
            }
        }
        if( (lookahead = get(input)) == EOF )
            return END_OF_FILE;
    }

    switch( lookahead )                  /* do the simple ones first */
    {
    case ']':   rval = CLOSE_BRACKET;   goto simple;
    case '[':   rval = OPEN_BRACKET;    goto simple;
    case '(':   rval = OPEN_PAREN;      goto simple;
    case ')':   rval = CLOSE_PAREN;     goto simple;
    case ':':   rval = COLON;           goto simple;
    case '-':   rval = DASH;            goto simple;
    case '!':   rval = NOT;             goto simple;

    simple:     yytext[0] = lookahead;
                yytext[1] = '\0';
                lookahead = -1;
                return rval;
    }

    if( isdigit( lookahead ) )                  /* Now do a NUMBER token */
    {
        p = yytext;
        while( isdigit(lookahead) )
        {
            *p++      = lookahead;
            lookahead = get( input );
        }
        *p = '\0';
        return NUMBER;
    }

    if( lookahead == '{' )                      /* A CODE token */
    {
        p = yytext;
        *p++ = '{';
        for(level = 1; level > 0 ;)
        {
            lookahead = get(input);
            *p++ = lookahead;

            if( lookahead == '{' )
                ++level;
            else if( lookahead == '}' )
                --level;
        }

        if( lookahead != '}' )
            error( "End of file encountered in accept block" );

        lookahead = -1;
        return CODE;
    }

    /* Other multiple-character tokens. Start by getting six characters of
     * lookahead (the length of the longest token) into the buffer, then set
     * lookahead to -1 to force a read on the next gtok call. The cmp()
     * calls match the lookahead buffer against possible tokens. If there's
     * a match, push back all characters in the lookahead buffer that aren't
     * part of the token, null terminate the lexem, then return the proper
     * token. If there's no match, its a CHARACTER token. Push back all but
     * the first character in this case.
     */

    p         = yytext;
    *p++      = lookahead;
    lookahead = -1;
    for( i = 5; --i >= 0; *p++ = get(input) )
        ;
    *p = '\0';
                                /* The numeric arguments to pushb() and the
                                 * following array indexes are both the length
                                 * of the template string.
                                 */
    if(!cmp("accept", yytext)) {                ; yytext[6]=0;  return ACCEPT; }
    if(!cmp("goto",   yytext)) { pushb(4,yytext); yytext[4]=0;  return GOTO;   }
    if(!cmp("loop",   yytext)) { pushb(4,yytext); yytext[4]=0;  return LOOP;   }
    if(!cmp("on",     yytext)) { pushb(2,yytext); yytext[2]=0;  return ON;     }
    if(!cmp("state",  yytext)) { pushb(5,yytext); yytext[5]=0;  return STATE;  }

    pushb(1, yytext);   /* push back all but the first character */
    yytext[1] = '\0';   /* terminate the string                  */
    return CHARACTER;   /* return it.                            */
}
